from dataclasses import dataclass
from enum import Enum
from typing import Sequence

import numpy as np
from numpy.typing import ArrayLike
from xarray.core.formatting import diff_attrs_repr


class DiffType(Enum):
    TAGS = 0
    PIXEL_STATS = 1,
    PIXEL = 2,
    DATASET = 3


DIFF_TYPE_PREFIXES = {
    DiffType.TAGS: "Differences in meta data:\n",
    DiffType.PIXEL_STATS: "Differences in pixel data:\n",
    DiffType.PIXEL: "Differences in pixels:\n",
    DiffType.DATASET: "Differences in dataset:\n",
}


@dataclass
class Difference:
    description: str
    type: DiffType


@dataclass
class Stats:
    min: float = 0
    max: float = 0
    mean: float = 0
    median: float = 0
    nans: int = 0

    def __str__(self):
        return ', '.join([f"{n}={getattr(self, n)}" for n in self.__dict__])

    @property
    def is_empty(self):
        return self.min == 0 and self.max == 0 and self.nans == 0


def calculate_pixel_diff_stats(approved_pixels: ArrayLike, received_pixels: ArrayLike) -> Stats:
    diff_pixels = np.abs(received_pixels - approved_pixels)
    if diff_pixels.size == 0:
        return Stats()
    diff_min = np.nanmin(diff_pixels).item()
    diff_max = np.nanmax(diff_pixels).item()
    diff_mean = np.nanmean(diff_pixels).item()
    diff_median = np.nanmedian(diff_pixels).item()
    diff_nans = (np.asarray(np.sum(np.isnan(received_pixels))) - np.asarray(np.sum(np.isnan(approved_pixels)))).item()
    return Stats(diff_min, diff_max, diff_mean, diff_median, diff_nans)


def print_diffs(diffs: Sequence[Difference]) -> None:
    for diff in diffs:
        print(f"{DIFF_TYPE_PREFIXES[diff.type]}{diff.description}")


def add_common_meta_data_diffs(a, b, diffs):
    a_data_vars = set(getattr(a, 'data_vars', {}))
    b_data_vars = set(getattr(b, 'data_vars', {}))
    common_data_vars = a_data_vars & b_data_vars
    a_coords = set(b.coords)
    b_coords = set(a.coords)
    common_coords = a_coords & b_coords
    common_sub_arrays = list(common_data_vars) + list(common_coords)
    diff_attrs = diff_attrs_repr(a.attrs, b.attrs, 'identical')
    if diff_attrs:
        diffs.append(Difference(diff_attrs, DiffType.TAGS))
    for name in common_sub_arrays:
        d = diff_attrs_repr(a[name].attrs, b[name].attrs, 'identical')
        if d:
            diffs.append(Difference(d, DiffType.TAGS))

    return diffs



from pathlib import Path
from typing import Sequence, Callable

import xarray as xr
from xarray import Dataset

from pytest_approvaltests_geo.differs.difference import Difference, add_common_meta_data_diffs, \
    calculate_pixel_diff_stats, DiffType
from pytest_approvaltests_geo.scrubbers import scrub_xarray_metadata, scrub_xarray_coordinates

DatasetOpener = Callable[[Path], Dataset]


class DifferOfGeoDataset:
    def __init__(self, opener: DatasetOpener, tags_scrubber, coords_scrubber, float_tolerance):
        self._opener = opener
        self._tags_scrubber = tags_scrubber
        self._coords_scrubber = coords_scrubber
        self._float_tolerance = float_tolerance

    def diffs(self, received_path: Path, approved_path: Path) -> Sequence[Difference]:
        diffs = []
        with self._opener(received_path) as received_ds, self._opener(approved_path) as approved_ds:
            received_ds = scrub_xarray_metadata(received_ds, self._tags_scrubber)
            approved_ds = scrub_xarray_metadata(approved_ds, self._tags_scrubber)
            received_ds = scrub_xarray_coordinates(received_ds, self._coords_scrubber)
            approved_ds = scrub_xarray_coordinates(approved_ds, self._coords_scrubber)
            diffs = add_common_meta_data_diffs(received_ds, approved_ds, diffs)

            try:
                xr.testing.assert_allclose(received_ds, approved_ds, **self._float_tolerance.to_kwargs())
            except AssertionError as assertion_diff:
                common_data_vars = set(received_ds.data_vars) & set(approved_ds.data_vars)
                stats_per_data_var = [(name, calculate_pixel_diff_stats(received_ds[name], approved_ds[name]))
                                      for name in common_data_vars]
                stats_per_data_var = [(n, s) for n, s in stats_per_data_var if not s.is_empty]
                diff_stats = '\n'.join([f"{n}: {str(s)}" for n, s in stats_per_data_var])
                if diff_stats:
                    diffs.append(Difference(diff_stats, DiffType.PIXEL_STATS))
                diffs.append(Difference(str(assertion_diff), DiffType.DATASET))

        return diffs



from contextlib import contextmanager
from pathlib import Path
from typing import Tuple, Dict

import rasterio
import rioxarray
from xarray import DataArray


@contextmanager
def read_array_and_tags(file_path: Path) -> Tuple[DataArray, Dict]:
    try:
        with rasterio.open(file_path) as rds:
            t = rds.tags()
            a = rioxarray.open_rasterio(rds)
        yield a, t
    finally:
        a.close()



from typing import Optional

import xarray as xr

from pytest_approvaltests_geo.differs.differ_of_geo_dataset import DifferOfGeoDataset
from pytest_approvaltests_geo.float_utils import Tolerance
from pytest_approvaltests_geo.scrubbers import RecursiveScrubber, identity_recursive_scrubber, SequenceScrubber, \
    identity_sequence_scrubber


class DifferOfGeoNcs(DifferOfGeoDataset):
    def __init__(self, tags_scrubber: Optional[RecursiveScrubber] = None,
                 coords_scrubber: Optional[SequenceScrubber] = None,
                 float_tolerance: Optional[Tolerance] = None):
        super().__init__(xr.open_dataset,
                         tags_scrubber or identity_recursive_scrubber,
                         coords_scrubber or identity_sequence_scrubber,
                         float_tolerance or Tolerance())



import shutil
from pathlib import Path

from approvaltests.core import Comparator

from pytest_approvaltests_geo.differs.differ_of_geo_zarrs import DifferOfGeoZarrs


class CompareGeoZarrs(Comparator, DifferOfGeoZarrs):

    def compare(self, received_path: str, approved_path: str) -> bool:
        received_path = Path(received_path)
        approved_path = Path(approved_path)
        if not received_path.exists() or not approved_path.exists():
            return False

        is_identical = len(self.diffs(received_path, approved_path)) == 0
        if is_identical:
            shutil.rmtree(received_path)
            received_path.touch()  # TODO: fix so approval tests has something to delete
        return is_identical



from pytest_approvaltests_geo.scrubbers import scrub_yeoda_datacube_metadata


def test_scrub_all_yeoda_metadata():
    assert scrub_yeoda_datacube_metadata("2018-09-12 12:00:00;20220101T235959;ae2df71;v1.0.2") == \
           "<date0>;<yeoda_date0>;<short_commit_0>;<tag_0>"



from pathlib import Path

import numpy as np
from approvaltests import Reporter
from xarray import Dataset

from pytest_approvaltests_geo.differs.differ_of_geo_zarrs import DifferOfGeoZarrs
from pytest_approvaltests_geo.differs.difference import print_diffs


class ReportGeoZarrs(Reporter, DifferOfGeoZarrs):
    def report(self, received_path: str, approved_path: str) -> bool:
        received_path = Path(received_path)
        approved_path = Path(approved_path)

        if not approved_path.exists():
            self._create_empty_ds(approved_path)

        diffs = self.diffs(received_path, approved_path)
        if len(diffs) > 0:
            print_diffs(diffs)
            print(f"To approve run:\nrm -rf {approved_path} && mv -f {received_path} {approved_path}")

        return True

    @staticmethod
    def _create_empty_ds(path: Path):
        Dataset({'empty': (("x", "y"), np.empty((1, 1)))}).to_zarr(path)



import numpy as np

from pytest_approvaltests_geo.differs.difference import calculate_pixel_diff_stats, Stats


def test_empty_arrays():
    assert calculate_pixel_diff_stats(np.array([]), np.array([])).is_empty


def test_all_nan_pixels():
    assert calculate_pixel_diff_stats(np.full((2, 2), np.nan), np.full((2, 2), np.nan)).nans == 0


def test_all_zero_pixels():
    assert calculate_pixel_diff_stats(np.zeros((2, 2)), np.zeros((2, 2))) == Stats()


def test_identical_pixels():
    assert calculate_pixel_diff_stats(np.array([[1, 2], [3, 4]]), np.array([[1, 2], [3, 4]])) == Stats()


def test_pixel_stats():
    stats = calculate_pixel_diff_stats(np.array([[1, 2], [3, 4]]), np.array([[2, 4], [4, 3]]))
    assert stats == Stats(min=1, max=2, mean=1.25, median=1.0, nans=0)
